@@ -58,6 +58,8 @@ our @EXPORT_OK = qw/
 	set_bKGD
 	get_cHRM
 	set_cHRM
+	get_eXIf
+	set_eXIf
 	get_gAMA
 	set_gAMA
 	get_hIST
@@ -99,7 +101,7 @@ our %EXPORT_TAGS = (
 );
 
 require XSLoader;
-our $VERSION = '0.49_01';
+our $VERSION = '0.49_02';
 
 XSLoader::load('Image::PNG::Libpng', $VERSION);
 
@@ -153,6 +155,8 @@ bKGD => 1,
 
 cHRM => 1,
 
+
+
 gAMA => 1,
 
 hIST => 1,
