@@ -1,4 +1,4 @@
-#line 2 "tmpl/perl-libpng.c.tmpl"
+#line 2 "perl-libpng.c.tmpl"
 
 /* Coding style note: use "color" in code & variables, not "colour",
    as per libpng. 2020-12-09 08:03:03 */
@@ -610,11 +610,11 @@ perl_png_set_text_from_hash (perl_libpng_t * png,
     SOFT_HASH_FETCH_PV (chunk, text);
     text_out->text = (char *) text;
     text_out->text_length = text_length;
-    /* Set this in case it starts to be required by future versions of
-       libpng. */
-    text_out->itxt_length = text_length;
 #ifdef PNG_iTXt_SUPPORTED
     if (is_itxt) {
+    /* Set this in case it starts to be required by future versions of
+       libpng. */
+	text_out->itxt_length = text_length;
 	SOFT_HASH_FETCH_PV (chunk, lang);
 	text_out->lang = (char *) lang;
 	SOFT_HASH_FETCH_PV (chunk, lang_key);
@@ -1270,18 +1270,19 @@ static void perl_png_set_bKGD (perl_libpng_t * png, HV * bKGD)
 
 static SV * perl_png_get_pCAL (perl_libpng_t * png)
 {
+    SV * pcal = & PL_sv_undef;
 #ifdef PNG_pCAL_SUPPORTED
     HV * ice;
     char * purpose;
-    int x0;
-    int x1;
+    png_int_32 x0;
+    png_int_32 x1;
     int type;
     int n_params;
     char * units;
     char ** png_params;
 
     if (! VALID (pCAL)) {
-	return & PL_sv_undef;
+	return pcal;
     }
     png_get_pCAL (pngi, & purpose, & x0, & x1, & type,
 		  & n_params, & units, & png_params);
@@ -1295,18 +1296,16 @@ static SV * perl_png_get_pCAL (perl_libpng_t * png)
 	AV * params;
 	int i;
 	params = newAV ();
-	/* Looks like a memory leak here. */
-	GET_MEMORY (png_params, n_params, char *);
 	for (i = 0; i < n_params; i++) {
 	    ARRAY_STORE_PV (params, png_params[i]);
 	}
 	HASH_STORE_AV (ice, params);
     }
-    return newRV_noinc ((SV *) ice);
+    pcal = newRV_noinc ((SV *) ice);
 #else
     UNSUPPORTED(pCAL);
-    return & PL_sv_undef;
 #endif
+    return pcal;
 }
 
 /* Set the pCAL (calibration of pixel values) chunk of a PNG
@@ -1345,6 +1344,9 @@ static void perl_png_set_pCAL (perl_libpng_t * png, HV * pCAL)
 	}
     }
     png_set_pCAL (pngi, purpose, x0, x1, type, n_params, units, png_params);
+    if (png_params) {
+	PERL_PNG_FREE (png_params);
+    }
 #else
     UNSUPPORTED(pCAL);
 #endif
@@ -1384,27 +1386,29 @@ static void perl_png_set_gAMA (perl_libpng_t * png, double gamma)
 
 static SV * perl_png_get_iCCP (perl_libpng_t * png)
 {
+    SV * iccp = & PL_sv_undef;
 #ifdef PNG_iCCP_SUPPORTED
-    if (VALID (iCCP)) {
-	char * name;
-	unsigned char * profile;
-	int compression_method;
-	unsigned int proflen;
-        HV * ice;
-	SV * profile_sv;
-	png_get_iCCP (pngi, & name, & compression_method, & profile,
-		      & proflen);
-        ice = newHV ();
-	HASH_STORE_PV (ice, name);
-	profile_sv = newSVpv ((char *) profile, proflen);
-	(void) hv_store (ice, "profile", strlen ("profile"), profile_sv, 0);
-        return newRV_noinc ((SV *) ice);
+    png_charp name;
+    png_bytep profile;
+    int compression_type = UNUSED_ZERO_ARG;
+    png_uint_32 proflen;
+    HV * ice;
+    SV * profile_sv;
+
+    if (! VALID (iCCP)) {
+	return iccp;
     }
-    return & PL_sv_undef;
+    png_get_iCCP (pngi, & name, & compression_type, & profile,
+		  & proflen);
+    ice = newHV ();
+    HASH_STORE_PV (ice, name);
+    profile_sv = newSVpv ((char *) profile, proflen);
+    (void) hv_store (ice, "profile", strlen ("profile"), profile_sv, 0);
+    iccp = newRV_noinc ((SV *) ice);
 #else /* PNG_iCCP_SUPPORTED */
     UNSUPPORTED(iCCP);
-    return & PL_sv_undef;
 #endif /* PNG_iCCP_SUPPORTED */
+    return iccp;
 }
 
 static void perl_png_set_iCCP (perl_libpng_t * png, HV * iCCP)
@@ -1414,15 +1418,12 @@ static void perl_png_set_iCCP (perl_libpng_t * png, HV * iCCP)
     STRLEN name_length;
     char * profile;
     STRLEN profile_length;
-    int compression_method;
-
-    compression_method = PNG_COMPRESSION_TYPE_BASE;
 
     HASH_FETCH_PV (iCCP, profile);
     HASH_FETCH_PV (iCCP, name);
 
-    png_set_iCCP (pngi, name, compression_method,
-		  (unsigned char *) profile, profile_length);
+    png_set_iCCP (pngi, name, UNUSED_ZERO_ARG,
+		  (png_bytep) profile, (png_uint_32) profile_length);
 #else /* PNG_iCCP_SUPPORTED */
     UNSUPPORTED(iCCP);
 #endif /* PNG_iCCP_SUPPORTED */
@@ -1466,7 +1467,7 @@ static SV * perl_png_get_tRNS (perl_libpng_t * png)
     else {
 	HV * trans_hv;
 	trans_hv = newHV ();
-#line 1469 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 1470 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
         
 	HASH_STORE_IV_MEMBER (trans_hv, red, trans_values);
 	
@@ -1476,7 +1477,7 @@ static SV * perl_png_get_tRNS (perl_libpng_t * png)
 	
 	HASH_STORE_IV_MEMBER (trans_hv, gray, trans_values);
 	
-#line 1474 "tmpl/perl-libpng.c.tmpl"
+#line 1475 "perl-libpng.c.tmpl"
 	return newRV_noinc ((SV *) trans_hv);
     }
 #else
@@ -1531,7 +1532,7 @@ static void perl_png_set_tRNS (perl_libpng_t * png, SV * tRNS)
 	    /* unhandled error */
 	}
 	trans_hv = (HV *) SvRV (tRNS);
-#line 1534 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 1535 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
         
 	    HASH_FETCH_IV_MEMBER (trans_hv, red, (& trans_values));
 	
@@ -1541,7 +1542,7 @@ static void perl_png_set_tRNS (perl_libpng_t * png, SV * tRNS)
 	
 	    HASH_FETCH_IV_MEMBER (trans_hv, gray, (& trans_values));
 	
-#line 1533 "tmpl/perl-libpng.c.tmpl"
+#line 1534 "perl-libpng.c.tmpl"
 	num_trans = 1;
 	png_set_tRNS (pngi, trans, num_trans, & trans_values); 
     }
@@ -1569,14 +1570,14 @@ perl_png_spalette_to_hv (png_sPLT_tp spalette)
 	png_sPLT_entry * entry;
 	entry = spalette->entries + i;
 	perl_entry = newHV ();
-#line 1572 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 1573 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
 		HASH_STORE_IV_MEMBER (perl_entry, red, entry);
 		HASH_STORE_IV_MEMBER (perl_entry, green, entry);
 		HASH_STORE_IV_MEMBER (perl_entry, blue, entry);
 		HASH_STORE_IV_MEMBER (perl_entry, alpha, entry);
 		HASH_STORE_IV_MEMBER (perl_entry, frequency, entry);
 	
-#line 1565 "tmpl/perl-libpng.c.tmpl"
+#line 1566 "perl-libpng.c.tmpl"
 	av_push (entries, newRV_noinc ((SV *) perl_entry));
     }
     HASH_STORE_AV (perl_spalette, entries);
@@ -1680,14 +1681,14 @@ static void perl_png_set_sPLT (perl_libpng_t * png, AV * sPLT_entries)
 	    }
 	    MESSAGE ("Copying entry %d", j);
 	    e = entry->entries + j;
-#line 1683 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 1684 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
  	    	    HASH_FETCH_IV_MEMBER (perl_entry, red, e);
 	    	    HASH_FETCH_IV_MEMBER (perl_entry, green, e);
 	    	    HASH_FETCH_IV_MEMBER (perl_entry, blue, e);
 	    	    HASH_FETCH_IV_MEMBER (perl_entry, alpha, e);
 	    	    HASH_FETCH_IV_MEMBER (perl_entry, frequency, e);
 	    
-#line 1673 "tmpl/perl-libpng.c.tmpl"
+#line 1674 "perl-libpng.c.tmpl"
 	}
     }
     MESSAGE ("Setting the entries");
@@ -1800,7 +1801,7 @@ static SV * perl_png_get_sBIT (perl_libpng_t * png)
 	if (status != PNG_INFO_sBIT) {
 	    /* error */
 	}
-#line 1803 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 1804 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
         
 	HASH_STORE_IV_MEMBER (sig_bit, red, colors);
 	
@@ -1812,7 +1813,7 @@ static SV * perl_png_get_sBIT (perl_libpng_t * png)
 	
 	HASH_STORE_IV_MEMBER (sig_bit, alpha, colors);
 	
-#line 1790 "tmpl/perl-libpng.c.tmpl"
+#line 1791 "perl-libpng.c.tmpl"
         return newRV_noinc ((SV *) sig_bit);
     }
     return & PL_sv_undef;
@@ -1994,9 +1995,9 @@ static SV * perl_png_get_valid (perl_libpng_t * png)
     valid = png_get_valid (pngi, 0xFFFFFFFF);
 #define V(x) \
     (void) hv_store (perl_valid, #x, strlen (#x), newSViv (valid & PNG_INFO_ ## x), 0)
-#line 1997 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 1998 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
 V(bKGD);V(cHRM);V(gAMA);V(hIST);V(iCCP);V(IDAT);V(oFFs);V(pCAL);V(pHYs);V(PLTE);V(sBIT);V(sCAL);V(sPLT);V(sRGB);V(tIME);V(tRNS);
-#line 1974 "tmpl/perl-libpng.c.tmpl"
+#line 1975 "perl-libpng.c.tmpl"
 #undef V
 
     return newRV_noinc ((SV *) perl_valid);
@@ -2370,7 +2371,7 @@ static void perl_png_set_unknown_chunks (perl_libpng_t * png, AV * chunk_list)
 
 int perl_png_libpng_supports (const char * what)
 {
-#line 2373 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 2374 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
     if (strcmp (what, "iTXt") == 0) {
 #ifdef PNG_iTXt_SUPPORTED
         return 1;
@@ -2539,7 +2540,35 @@ int perl_png_libpng_supports (const char * what)
         return 0;
 #endif /* WRITE_CUSTOMIZE_COMPRESSION */
     }
-#line 2362 "tmpl/perl-libpng.c.tmpl"
+    if (strcmp (what, "eXIf") == 0) {
+#ifdef PNG_eXIf_SUPPORTED
+        return 1;
+#else
+        return 0;
+#endif /* eXIf */
+    }
+    if (strcmp (what, "gIFg") == 0) {
+#ifdef PNG_gIFg_SUPPORTED
+        return 1;
+#else
+        return 0;
+#endif /* gIFg */
+    }
+    if (strcmp (what, "gIFx") == 0) {
+#ifdef PNG_gIFx_SUPPORTED
+        return 1;
+#else
+        return 0;
+#endif /* gIFx */
+    }
+    if (strcmp (what, "sTER") == 0) {
+#ifdef PNG_sTER_SUPPORTED
+        return 1;
+#else
+        return 0;
+#endif /* sTER */
+    }
+#line 2366 "perl-libpng.c.tmpl"
 
     /* sCAL is a special case. */
 
@@ -2645,7 +2674,7 @@ static SV * perl_png_get_cHRM (perl_libpng_t * png)
 #ifdef PNG_cHRM_SUPPORTED
     if (VALID (cHRM)) {
         HV * ice;
-#line 2648 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 2677 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
         double white_x;
         double white_y;
         double red_x;
@@ -2654,10 +2683,10 @@ static SV * perl_png_get_cHRM (perl_libpng_t * png)
         double green_y;
         double blue_x;
         double blue_y;
-#line 2472 "tmpl/perl-libpng.c.tmpl"
+#line 2476 "perl-libpng.c.tmpl"
         png_get_cHRM (pngi , & white_x, & white_y, & red_x, & red_y, & green_x, & green_y, & blue_x, & blue_y);
         ice = newHV ();
-#line 2660 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 2689 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
         (void) hv_store (ice, "white_x", strlen ("white_x"),
                          newSVnv (white_x), 0);
         (void) hv_store (ice, "white_y", strlen ("white_y"),
@@ -2674,7 +2703,7 @@ static SV * perl_png_get_cHRM (perl_libpng_t * png)
                          newSVnv (blue_x), 0);
         (void) hv_store (ice, "blue_y", strlen ("blue_y"),
                          newSVnv (blue_y), 0);
-#line 2481 "tmpl/perl-libpng.c.tmpl"
+#line 2485 "perl-libpng.c.tmpl"
         return newRV_noinc ((SV *) ice);
     }
     return & PL_sv_undef;
@@ -2689,7 +2718,7 @@ static void perl_png_set_cHRM (perl_libpng_t * png, HV * cHRM)
 {
 #ifdef PNG_cHRM_SUPPORTED
     SV ** key_sv_ptr;
-#line 2692 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 2721 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
     double white_x = 0.0;
     double white_y = 0.0;
     double red_x = 0.0;
@@ -2730,9 +2759,9 @@ static void perl_png_set_cHRM (perl_libpng_t * png, HV * cHRM)
     if (key_sv_ptr) {
         blue_y = SvNV (* key_sv_ptr);
     }
-#line 2506 "tmpl/perl-libpng.c.tmpl"
+#line 2510 "perl-libpng.c.tmpl"
     png_set_cHRM (pngi, white_x, white_y, red_x, red_y, green_x, green_y, blue_x, blue_y);
-#line 2509 "tmpl/perl-libpng.c.tmpl"
+#line 2513 "perl-libpng.c.tmpl"
 #else
     /* libpng was compiled without this option. */
     UNSUPPORTED(cHRM);
@@ -2766,7 +2795,7 @@ static void perl_png_copy_row_pointers (perl_libpng_t * png, SV * row_pointers)
     png_set_rows (pngi, png->row_pointers);
 }
 
-#line 2769 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 2798 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
 
 static int
 perl_png_get_image_width (perl_libpng_t * png)
@@ -2780,7 +2809,7 @@ perl_png_get_image_height (perl_libpng_t * png)
     return png_get_image_height (pngi);
 }
 
-#line 2551 "tmpl/perl-libpng.c.tmpl"
+#line 2555 "perl-libpng.c.tmpl"
 
 
 static void
@@ -2841,6 +2870,41 @@ perl_png_get_user_height_max (perl_libpng_t * png)
 }
 
 
+static SV *
+perl_png_get_eXIf (perl_libpng_t * png)
+{
+    SV * exif = & PL_sv_undef;
+#ifdef PNG_eXIf_SUPPORTED
+    if (VALID (eXIf)) {
+	png_uint_32 num_exif;
+	png_bytep exif_buf;
+	
+	png_get_eXIf_1 (png->png, png->info, & num_exif, & exif_buf);
+	exif = newSVpvn ((const char *) exif_buf, (STRLEN) num_exif);
+    }
+#else /*  PNG_eXIf_SUPPORTED */
+    UNSUPPORTED (eXIf);
+#endif /*  PNG_eXIf_SUPPORTED */
+    return exif;
+}
+
+static void
+perl_png_set_eXIf (perl_libpng_t * png, SV * exif)
+{
+#ifdef PNG_eXIf_SUPPORTED
+    png_uint_32 num_exif;
+    png_bytep exif_buf;
+    STRLEN exif_len;
+    
+    exif_buf = (png_bytep) SvPV (exif, exif_len);
+    num_exif = (png_uint_32) exif_len;
+
+    png_set_eXIf_1 (png->png, png->info, num_exif, exif_buf);
+#else /*  PNG_eXIf_SUPPORTED */
+    UNSUPPORTED (eXIf);
+#endif /*  PNG_eXIf_SUPPORTED */
+}
+
 static SV *
 perl_png_split_alpha (perl_libpng_t * png)
 {
